@@ -25,6 +25,7 @@
 #include <cstdlib>
 #include <cassert>
 #include <cerrno>
+#include <cstring>
 #include <iostream>
 #include <algorithm>
 #include <memory>
@@ -184,8 +185,6 @@ Client::Client(struct ev_loop *loop, uint32_t client_chosen_version,
       client_chosen_version_(client_chosen_version),
       original_version_(original_version),
       early_data_(false),
-      should_exit_(false),
-      should_exit_on_handshake_confirmed_(false),
       handshake_confirmed_(false),
       tx_{} {
   ev_io_init(&wev_, writecb, 0, EV_WRITE);
@@ -232,14 +231,14 @@ void Client::disconnect() {
 }
 
 namespace {
-int recv_crypto_data(ngtcp2_conn *conn, ngtcp2_crypto_level crypto_level,
-                     uint64_t offset, const uint8_t *data, size_t datalen,
-                     void *user_data) {
+int recv_crypto_data(ngtcp2_conn *conn,
+                     ngtcp2_encryption_level encryption_level, uint64_t offset,
+                     const uint8_t *data, size_t datalen, void *user_data) {
   if (!config.quiet && !config.no_quic_dump) {
-    debug::print_crypto_data(crypto_level, data, datalen);
+    debug::print_crypto_data(encryption_level, data, datalen);
   }
 
-  return ngtcp2_crypto_recv_crypto_data_cb(conn, crypto_level, offset, data,
+  return ngtcp2_crypto_recv_crypto_data_cb(conn, encryption_level, offset, data,
                                            datalen, user_data);
 }
 } // namespace
@@ -300,9 +299,9 @@ int Client::handshake_completed() {
     // handshake completion (e.g., OpenSSL).  For TLS backends which
     // report it early (e.g., BoringSSL and PicoTLS), the following
     // functions are noop.
-    if (auto rv = ngtcp2_conn_early_data_rejected(conn_); rv != 0) {
-      std::cerr << "ngtcp2_conn_early_data_rejected: " << ngtcp2_strerror(rv)
-                << std::endl;
+    if (auto rv = ngtcp2_conn_tls_early_data_rejected(conn_); rv != 0) {
+      std::cerr << "ngtcp2_conn_tls_early_data_rejected: "
+                << ngtcp2_strerror(rv) << std::endl;
       return -1;
     }
   }
@@ -315,9 +314,14 @@ int Client::handshake_completed() {
   }
 
   if (config.tp_file) {
-    auto params = ngtcp2_conn_get_remote_transport_params(conn_);
-
-    if (write_transport_params(config.tp_file, params) != 0) {
+    std::array<uint8_t, 256> data;
+    auto datalen = ngtcp2_conn_encode_0rtt_transport_params(conn_, data.data(),
+                                                            data.size());
+    if (datalen < 0) {
+      std::cerr << "Could not encode 0-RTT transport parameters: "
+                << ngtcp2_strerror(datalen) << std::endl;
+    } else if (util::write_transport_params(config.tp_file, data.data(),
+                                            datalen) != 0) {
       std::cerr << "Could not write transport parameters in " << config.tp_file
                 << std::endl;
     }
@@ -342,6 +346,16 @@ int handshake_confirmed(ngtcp2_conn *conn, void *user_data) {
 }
 } // namespace
 
+bool Client::should_exit() const {
+  return handshake_confirmed_ &&
+         (!config.wait_for_ticket || ticket_received_) &&
+         ((config.exit_on_first_stream_close &&
+           (config.nstreams == 0 || nstreams_closed_)) ||
+          (config.exit_on_all_streams_close &&
+           config.nstreams == nstreams_done_ &&
+           nstreams_closed_ == nstreams_done_));
+}
+
 int Client::handshake_confirmed() {
   handshake_confirmed_ = true;
 
@@ -355,10 +369,6 @@ int Client::handshake_confirmed() {
     start_delay_stream_timer();
   }
 
-  if (should_exit_on_handshake_confirmed_) {
-    should_exit_ = true;
-  }
-
   return 0;
 }
 
@@ -407,7 +417,7 @@ int extend_max_streams_bidi(ngtcp2_conn *conn, uint64_t max_streams,
 
 namespace {
 void rand(uint8_t *dest, size_t destlen, const ngtcp2_rand_ctx *rand_ctx) {
-  auto dis = std::uniform_int_distribution<uint8_t>(0, 255);
+  auto dis = std::uniform_int_distribution<uint8_t>();
   std::generate(dest, dest + destlen, [&dis]() { return dis(randgen); });
 }
 } // namespace
@@ -466,6 +476,7 @@ int update_key(ngtcp2_conn *conn, uint8_t *rx_secret, uint8_t *tx_secret,
 
 namespace {
 int path_validation(ngtcp2_conn *conn, uint32_t flags, const ngtcp2_path *path,
+                    const ngtcp2_path *old_path,
                     ngtcp2_path_validation_result res, void *user_data) {
   if (!config.quiet) {
     debug::path_validation(path, res);
@@ -539,7 +550,7 @@ int Client::extend_max_stream_data(int64_t stream_id, uint64_t max_data) {
 }
 
 namespace {
-int recv_new_token(ngtcp2_conn *conn, const ngtcp2_vec *token,
+int recv_new_token(ngtcp2_conn *conn, const uint8_t *token, size_t tokenlen,
                    void *user_data) {
   if (config.token_file.empty()) {
     return 0;
@@ -551,7 +562,7 @@ int recv_new_token(ngtcp2_conn *conn, const ngtcp2_vec *token,
     return 0;
   }
 
-  PEM_write_bio(f, "QUIC TOKEN", "", token->base, token->len);
+  PEM_write_bio(f, "QUIC TOKEN", "", token, tokenlen);
   BIO_free(f);
 
   return 0;
@@ -668,11 +679,11 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
                 << strerror(errno) << std::endl;
       return -1;
     }
-    settings.qlog.write = qlog_write_cb;
+    settings.qlog_write = qlog_write_cb;
   }
 
   settings.cc_algo = config.cc_algo;
-  settings.initial_ts = util::timestamp(loop_);
+  settings.initial_ts = util::timestamp();
   settings.initial_rtt = config.initial_rtt;
   settings.max_window = config.max_window;
   settings.max_stream_window = config.max_stream_window;
@@ -683,6 +694,12 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
   settings.handshake_timeout = config.handshake_timeout;
   settings.no_pmtud = config.no_pmtud;
   settings.ack_thresh = config.ack_thresh;
+  if (config.initial_pkt_num == UINT32_MAX) {
+    auto dis = std::uniform_int_distribution<uint32_t>(0, INT32_MAX);
+    settings.initial_pkt_num = dis(randgen);
+  } else {
+    settings.initial_pkt_num = config.initial_pkt_num;
+  }
 
   std::string token;
 
@@ -692,14 +709,14 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
     auto t = util::read_token(config.token_file);
     if (t) {
       token = std::move(*t);
-      settings.token.base = reinterpret_cast<uint8_t *>(token.data());
-      settings.token.len = token.size();
+      settings.token = reinterpret_cast<const uint8_t *>(token.data());
+      settings.tokenlen = token.size();
     }
   }
 
-  if (!config.other_versions.empty()) {
-    settings.other_versions = config.other_versions.data();
-    settings.other_versionslen = config.other_versions.size();
+  if (!config.available_versions.empty()) {
+    settings.available_versions = config.available_versions.data();
+    settings.available_versionslen = config.available_versions.size();
   }
 
   if (!config.preferred_versions.empty()) {
@@ -749,14 +766,18 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
   ngtcp2_conn_set_tls_native_handle(conn_, tls_session_.get_native_handle());
 
   if (early_data_ && config.tp_file) {
-    ngtcp2_transport_params params;
-    if (read_transport_params(config.tp_file, &params) != 0) {
-      std::cerr << "Could not read transport parameters from " << config.tp_file
-                << std::endl;
+    auto params = util::read_transport_params(config.tp_file);
+    if (!params) {
       early_data_ = false;
     } else {
-      ngtcp2_conn_set_early_remote_transport_params(conn_, &params);
-      if (make_stream_early() != 0) {
+      auto rv = ngtcp2_conn_decode_and_set_0rtt_transport_params(
+          conn_, reinterpret_cast<const uint8_t *>(params->data()),
+          params->size());
+      if (rv != 0) {
+        std::cerr << "ngtcp2_conn_decode_and_set_0rtt_transport_params:"
+                  << ngtcp2_strerror(rv) << std::endl;
+        early_data_ = false;
+      } else if (make_stream_early() != 0) {
         return -1;
       }
     }
@@ -772,6 +793,10 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
 int Client::feed_data(const Endpoint &ep, const sockaddr *sa, socklen_t salen,
                       const ngtcp2_pkt_info *pi, uint8_t *data,
                       size_t datalen) {
+  if (datalen == 0) {
+    return 0;
+  }
+
   auto path = ngtcp2_path{
       {
           const_cast<sockaddr *>(&ep.addr.su.sa),
@@ -784,16 +809,15 @@ int Client::feed_data(const Endpoint &ep, const sockaddr *sa, socklen_t salen,
       const_cast<Endpoint *>(&ep),
   };
   if (auto rv = ngtcp2_conn_read_pkt(conn_, &path, pi, data, datalen,
-                                     util::timestamp(loop_));
+                                     util::timestamp());
       rv != 0) {
     std::cerr << "ngtcp2_conn_read_pkt: " << ngtcp2_strerror(rv) << std::endl;
     if (!last_error_.error_code) {
       if (rv == NGTCP2_ERR_CRYPTO) {
-        ngtcp2_connection_close_error_set_transport_error_tls_alert(
+        ngtcp2_ccerr_set_tls_alert(
             &last_error_, ngtcp2_conn_get_tls_alert(conn_), nullptr, 0);
       } else {
-        ngtcp2_connection_close_error_set_transport_error_liberr(
-            &last_error_, rv, nullptr, 0);
+        ngtcp2_ccerr_set_liberr(&last_error_, rv, nullptr, 0);
       }
     }
     disconnect();
@@ -803,7 +827,7 @@ int Client::feed_data(const Endpoint &ep, const sockaddr *sa, socklen_t salen,
 }
 
 int Client::on_read(const Endpoint &ep) {
-  std::array<uint8_t, 64_k> buf;
+  std::array<uint8_t, 64_k - 1> buf;
   sockaddr_union su;
   size_t pktcnt = 0;
   ngtcp2_pkt_info pi;
@@ -817,7 +841,7 @@ int Client::on_read(const Endpoint &ep) {
   msg.msg_iov = &msg_iov;
   msg.msg_iovlen = 1;
 
-  uint8_t msg_ctrl[CMSG_SPACE(sizeof(uint8_t))];
+  uint8_t msg_ctrl[CMSG_SPACE(sizeof(uint8_t)) + CMSG_SPACE(sizeof(uint16_t))];
   msg.msg_control = msg_ctrl;
 
   for (;;) {
@@ -834,32 +858,48 @@ int Client::on_read(const Endpoint &ep) {
     }
 
     pi.ecn = msghdr_get_ecn(&msg, su.storage.ss_family);
-
-    if (!config.quiet) {
-      std::cerr << "Received packet: local="
-                << util::straddr(&ep.addr.su.sa, ep.addr.len)
-                << " remote=" << util::straddr(&su.sa, msg.msg_namelen)
-                << " ecn=0x" << std::hex << pi.ecn << std::dec << " " << nread
-                << " bytes" << std::endl;
+    auto gso_size = msghdr_get_udp_gro(&msg);
+    if (gso_size == 0) {
+      gso_size = static_cast<size_t>(nread);
     }
 
-    if (debug::packet_lost(config.rx_loss_prob)) {
+    auto data = buf.data();
+
+    for (;;) {
+      auto datalen = std::min(static_cast<size_t>(nread), gso_size);
+
+      ++pktcnt;
+
       if (!config.quiet) {
-        std::cerr << "** Simulated incoming packet loss **" << std::endl;
+        std::cerr << "Received packet: local="
+                  << util::straddr(&ep.addr.su.sa, ep.addr.len)
+                  << " remote=" << util::straddr(&su.sa, msg.msg_namelen)
+                  << " ecn=0x" << std::hex << pi.ecn << std::dec << " "
+                  << datalen << " bytes" << std::endl;
       }
-      break;
+
+      if (debug::packet_lost(config.rx_loss_prob)) {
+        if (!config.quiet) {
+          std::cerr << "** Simulated incoming packet loss **" << std::endl;
+        }
+      } else {
+        feed_data(ep, &su.sa, msg.msg_namelen, &pi, data, datalen);
+      }
+
+      nread -= datalen;
+      if (nread == 0) {
+        break;
+      }
+
+      data += datalen;
     }
 
-    if (feed_data(ep, &su.sa, msg.msg_namelen, &pi, buf.data(), nread) != 0) {
-      return -1;
-    }
-
-    if (++pktcnt >= 10) {
+    if (pktcnt >= 10) {
       break;
     }
   }
 
-  if (should_exit_) {
+  if (should_exit()) {
     disconnect();
     return -1;
   }
@@ -870,12 +910,11 @@ int Client::on_read(const Endpoint &ep) {
 }
 
 int Client::handle_expiry() {
-  auto now = util::timestamp(loop_);
+  auto now = util::timestamp();
   if (auto rv = ngtcp2_conn_handle_expiry(conn_, now); rv != 0) {
     std::cerr << "ngtcp2_conn_handle_expiry: " << ngtcp2_strerror(rv)
               << std::endl;
-    ngtcp2_connection_close_error_set_transport_error_liberr(&last_error_, rv,
-                                                             nullptr, 0);
+    ngtcp2_ccerr_set_liberr(&last_error_, rv, nullptr, 0);
     disconnect();
     return -1;
   }
@@ -892,13 +931,15 @@ int Client::on_write() {
     if (tx_.send_blocked) {
       return 0;
     }
+
+    ev_io_stop(loop_, &wev_);
   }
 
   if (auto rv = write_streams(); rv != 0) {
     return rv;
   }
 
-  if (should_exit_) {
+  if (should_exit()) {
     disconnect();
     return -1;
   }
@@ -913,7 +954,7 @@ int Client::write_streams() {
   size_t pktcnt = 0;
   auto max_udp_payload_size = ngtcp2_conn_get_max_tx_udp_payload_size(conn_);
   auto max_pktcnt = ngtcp2_conn_get_send_quantum(conn_) / max_udp_payload_size;
-  auto ts = util::timestamp(loop_);
+  auto ts = util::timestamp();
 
   ngtcp2_path_storage_zero(&ps);
 
@@ -959,8 +1000,7 @@ int Client::write_streams() {
 
       std::cerr << "ngtcp2_conn_write_stream: " << ngtcp2_strerror(nwrite)
                 << std::endl;
-      ngtcp2_connection_close_error_set_transport_error_liberr(
-          &last_error_, nwrite, nullptr, 0);
+      ngtcp2_ccerr_set_liberr(&last_error_, nwrite, nullptr, 0);
       disconnect();
       return -1;
     } else if (ndatalen >= 0) {
@@ -973,7 +1013,6 @@ int Client::write_streams() {
     if (nwrite == 0) {
       // We are congestion limited.
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      ev_io_stop(loop_, &wev_);
       return 0;
     }
 
@@ -983,8 +1022,7 @@ int Client::write_streams() {
             send_packet(ep, ps.path.remote, pi.ecn, tx_.data.data(), nwrite);
         rv != NETWORK_ERR_OK) {
       if (rv != NETWORK_ERR_SEND_BLOCKED) {
-        ngtcp2_connection_close_error_set_transport_error_liberr(
-            &last_error_, NGTCP2_ERR_INTERNAL, nullptr, 0);
+        ngtcp2_ccerr_set_liberr(&last_error_, NGTCP2_ERR_INTERNAL, nullptr, 0);
         disconnect();
 
         return rv;
@@ -998,7 +1036,6 @@ int Client::write_streams() {
 
     if (++pktcnt == max_pktcnt) {
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      start_wev_endpoint(ep);
       return 0;
     }
   }
@@ -1006,12 +1043,13 @@ int Client::write_streams() {
 
 void Client::update_timer() {
   auto expiry = ngtcp2_conn_get_expiry(conn_);
-  auto now = util::timestamp(loop_);
+  auto now = util::timestamp();
 
   if (expiry <= now) {
     if (!config.quiet) {
       auto t = static_cast<ev_tstamp>(now - expiry) / NGTCP2_SECONDS;
-      std::cerr << "Timer has already expired: " << t << "s" << std::endl;
+      std::cerr << "Timer has already expired: " << std::fixed << t << "s"
+                << std::defaultfloat << std::endl;
     }
 
     ev_feed_event(loop_, &timer_, EV_TIMER);
@@ -1115,6 +1153,7 @@ int udp_sock(int family) {
   fd_set_recv_ecn(fd, family);
   fd_set_ip_mtu_discover(fd, family);
   fd_set_ip_dontfrag(fd, family);
+  fd_set_udp_gro(fd);
 
   return fd;
 }
@@ -1271,8 +1310,8 @@ int Client::change_local_addr() {
         },
         &ep,
     };
-    if (auto rv = ngtcp2_conn_initiate_immediate_migration(
-            conn_, &path, util::timestamp(loop_));
+    if (auto rv = ngtcp2_conn_initiate_immediate_migration(conn_, &path,
+                                                           util::timestamp());
         rv != 0) {
       std::cerr << "ngtcp2_conn_initiate_immediate_migration: "
                 << ngtcp2_strerror(rv) << std::endl;
@@ -1330,7 +1369,7 @@ int Client::initiate_key_update() {
     std::cerr << "Initiate key update" << std::endl;
   }
 
-  if (auto rv = ngtcp2_conn_initiate_key_update(conn_, util::timestamp(loop_));
+  if (auto rv = ngtcp2_conn_initiate_key_update(conn_, util::timestamp());
       rv != 0) {
     std::cerr << "ngtcp2_conn_initiate_key_update: " << ngtcp2_strerror(rv)
               << std::endl;
@@ -1440,13 +1479,10 @@ int Client::send_blocked_packet() {
     if (rv == NETWORK_ERR_SEND_BLOCKED) {
       assert(wev_.fd == tx_.blocked.endpoint->fd);
 
-      ev_io_start(loop_, &wev_);
-
       return 0;
     }
 
-    ngtcp2_connection_close_error_set_transport_error_liberr(
-        &last_error_, NGTCP2_ERR_INTERNAL, nullptr, 0);
+    ngtcp2_ccerr_set_liberr(&last_error_, NGTCP2_ERR_INTERNAL, nullptr, 0);
     disconnect();
 
     return rv;
@@ -1458,8 +1494,8 @@ int Client::send_blocked_packet() {
 }
 
 int Client::handle_error() {
-  if (!conn_ || ngtcp2_conn_is_in_closing_period(conn_) ||
-      ngtcp2_conn_is_in_draining_period(conn_)) {
+  if (!conn_ || ngtcp2_conn_in_closing_period(conn_) ||
+      ngtcp2_conn_in_draining_period(conn_)) {
     return 0;
   }
 
@@ -1473,7 +1509,7 @@ int Client::handle_error() {
 
   auto nwrite = ngtcp2_conn_write_connection_close(
       conn_, &ps.path, &pi, buf.data(), buf.size(), &last_error_,
-      util::timestamp(loop_));
+      util::timestamp());
   if (nwrite < 0) {
     std::cerr << "ngtcp2_conn_write_connection_close: "
               << ngtcp2_strerror(nwrite) << std::endl;
@@ -1497,16 +1533,6 @@ int Client::on_stream_close(int64_t stream_id, uint64_t app_error_code) {
 
   ++nstreams_closed_;
 
-  if (config.exit_on_first_stream_close ||
-      (config.exit_on_all_streams_close && config.nstreams == nstreams_done_ &&
-       nstreams_closed_ == nstreams_done_)) {
-    if (handshake_confirmed_) {
-      should_exit_ = true;
-    } else {
-      should_exit_on_handshake_confirmed_ = true;
-    }
-  }
-
   if (!ngtcp2_is_bidi_stream(stream_id)) {
     assert(!ngtcp2_conn_is_local_stream(conn_, stream_id));
     ngtcp2_conn_extend_max_streams_uni(conn_, 1);
@@ -1796,8 +1822,9 @@ void config_set_default(Config &config) {
   config.max_streams_uni = 100;
   config.cc_algo = NGTCP2_CC_ALGO_CUBIC;
   config.initial_rtt = NGTCP2_DEFAULT_INITIAL_RTT;
-  config.handshake_timeout = NGTCP2_DEFAULT_HANDSHAKE_TIMEOUT;
+  config.handshake_timeout = UINT64_MAX;
   config.ack_thresh = 2;
+  config.initial_pkt_num = UINT32_MAX;
 }
 } // namespace
 
@@ -1833,8 +1860,7 @@ Options:
               version is  not supported by libngtcp2,  client will use
               QUIC v1  long packet  types.  Instead of  specifying hex
               string,  there  are   special  aliases  available:  "v1"
-              indicates  QUIC  v1,  and "v2draft"  indicates  QUIC  v2
-              draft.
+              indicates QUIC v1, and "v2" indicates QUIC v2.
               Default: )"
             << std::hex << "0x" << config.version << std::dec << R"(
   --preferred-versions=<HEX>[[,<HEX>]...]
@@ -1843,16 +1869,16 @@ Options:
               client received Version  Negotiation packet from server.
               These versions must be  supported by libngtcp2.  Instead
               of  specifying hex  string,  there  are special  aliases
-              available:  "v1"   indicates  QUIC  v1,   and  "v2draft"
-              indicates QUIC v2 draft.
-  --other-versions=<HEX>[[,<HEX>]...]
+              available: "v1"  indicates QUIC  v1, and  "v2" indicates
+              QUIC v2.
+  --available-versions=<HEX>[[,<HEX>]...]
               Specify QUIC  versions in  hex string  that are  sent in
-              other_versions  field  of version_information  transport
-              parameter.  This list can include a version which is not
-              supported  by  libngtcp2.   Instead  of  specifying  hex
-              string,  there  are   special  aliases  available:  "v1"
-              indicates  QUIC  v1,  and "v2draft"  indicates  QUIC  v2
-              draft.
+              available_versions    field    of    version_information
+              transport parameter.   This list  can include  a version
+              which  is  not  supported   by  libngtcp2.   Instead  of
+              specifying  hex   string,  there  are   special  aliases
+              available: "v1"  indicates QUIC  v1, and  "v2" indicates
+              QUIC v2.
   -q, --quiet Suppress debug output.
   -s, --show-secret
               Print out secrets unless --quiet is used.
@@ -1948,9 +1974,15 @@ Options:
               Default: )"
             << config.max_streams_uni << R"(
   --exit-on-first-stream-close
-              Exit when a first HTTP stream is closed.
+              Exit  when  a  first  client initiated  HTTP  stream  is
+              closed.
   --exit-on-all-streams-close
-              Exit when all HTTP streams are closed.
+              Exit when all client initiated HTTP streams are closed.
+  --wait-for-ticket
+              Wait  for a  ticket  to be  received  before exiting  on
+              --exit-on-first-stream-close                          or
+              --exit-on-all-streams-close.   --session-file   must  be
+              specified.
   --disable-early-data
               Disable early data.
   --cc=(cubic|reno|bbr|bbr2)
@@ -1982,15 +2014,19 @@ Options:
   --max-udp-payload-size=<SIZE>
               Override maximum UDP payload size that client transmits.
   --handshake-timeout=<DURATION>
-              Set the QUIC handshake timeout.
-              Default: )"
-            << util::format_duration(config.handshake_timeout) << R"(
+              Set  the  QUIC handshake  timeout.   It  defaults to  no
+              timeout.
   --no-pmtud  Disables Path MTU Discovery.
   --ack-thresh=<N>
               The minimum number of the received ACK eliciting packets
               that triggers immediate acknowledgement.
               Default: )"
             << config.ack_thresh << R"(
+  --initial-pkt-num=<N>
+              The initial packet  number that is used  for each packet
+              number space.  It  must be in range [0, (1  << 31) - 1],
+              inclusive.   By default,  the initial  packet number  is
+              chosen randomly.
   -h, --help  Display this help and exit.
 
 ---
@@ -2063,10 +2099,12 @@ int main(int argc, char **argv) {
         {"max-stream-window", required_argument, &flag, 33},
         {"max-udp-payload-size", required_argument, &flag, 35},
         {"handshake-timeout", required_argument, &flag, 36},
-        {"other-versions", required_argument, &flag, 37},
+        {"available-versions", required_argument, &flag, 37},
         {"no-pmtud", no_argument, &flag, 38},
         {"preferred-versions", required_argument, &flag, 39},
         {"ack-thresh", required_argument, &flag, 40},
+        {"wait-for-ticket", no_argument, &flag, 41},
+        {"initial-pkt-num", required_argument, &flag, 42},
         {nullptr, 0, nullptr, 0},
     };
 
@@ -2123,8 +2161,8 @@ int main(int argc, char **argv) {
         config.version = NGTCP2_PROTO_VER_V1;
         break;
       }
-      if (optarg == "v2draft"sv) {
-        config.version = NGTCP2_PROTO_VER_V2_DRAFT;
+      if (optarg == "v2"sv) {
+        config.version = NGTCP2_PROTO_VER_V2;
         break;
       }
       auto rv = util::parse_version(optarg);
@@ -2390,27 +2428,27 @@ int main(int argc, char **argv) {
         }
         break;
       case 37: {
-        // --other-versions
+        // --available-versions
         if (strlen(optarg) == 0) {
-          config.other_versions.resize(0);
+          config.available_versions.resize(0);
           break;
         }
         auto l = util::split_str(optarg);
-        config.other_versions.resize(l.size());
-        auto it = std::begin(config.other_versions);
+        config.available_versions.resize(l.size());
+        auto it = std::begin(config.available_versions);
         for (const auto &k : l) {
           if (k == "v1"sv) {
             *it++ = NGTCP2_PROTO_VER_V1;
             continue;
           }
-          if (k == "v2draft"sv) {
-            *it++ = NGTCP2_PROTO_VER_V2_DRAFT;
+          if (k == "v2"sv) {
+            *it++ = NGTCP2_PROTO_VER_V2;
             continue;
           }
           auto rv = util::parse_version(k);
           if (!rv) {
-            std::cerr << "other-versions: invalid version " << std::quoted(k)
-                      << std::endl;
+            std::cerr << "available-versions: invalid version "
+                      << std::quoted(k) << std::endl;
             exit(EXIT_FAILURE);
           }
           *it++ = *rv;
@@ -2435,8 +2473,8 @@ int main(int argc, char **argv) {
             *it++ = NGTCP2_PROTO_VER_V1;
             continue;
           }
-          if (k == "v2draft"sv) {
-            *it++ = NGTCP2_PROTO_VER_V2_DRAFT;
+          if (k == "v2"sv) {
+            *it++ = NGTCP2_PROTO_VER_V2;
             continue;
           }
           auto rv = util::parse_version(k);
@@ -2466,6 +2504,23 @@ int main(int argc, char **argv) {
           config.ack_thresh = *n;
         }
         break;
+      case 41:
+        // --wait-for-ticket
+        config.wait_for_ticket = true;
+        break;
+      case 42:
+        // --initial-pkt-num
+        if (auto n = util::parse_uint(optarg); !n) {
+          std::cerr << "initial-pkt-num: invalid argument" << std::endl;
+          exit(EXIT_FAILURE);
+        } else if (*n > INT32_MAX) {
+          std::cerr << "initial-pkt-num: must not exceed (1 << 31) - 1"
+                    << std::endl;
+          exit(EXIT_FAILURE);
+        } else {
+          config.initial_pkt_num = static_cast<uint32_t>(*n);
+        }
+        break;
       }
       break;
     default:
@@ -2491,6 +2546,11 @@ int main(int argc, char **argv) {
     exit(EXIT_FAILURE);
   }
 
+  if (config.wait_for_ticket && !config.session_file) {
+    std::cerr << "wait-for-ticket: session-file must be specified" << std::endl;
+    exit(EXIT_FAILURE);
+  }
+
   if (data_path) {
     auto fd = open(data_path, O_RDONLY);
     if (fd == -1) {
@@ -2527,17 +2587,17 @@ int main(int argc, char **argv) {
         std::find(std::begin(config.preferred_versions),
                   std::end(config.preferred_versions),
                   config.version) == std::end(config.preferred_versions)) {
-      std::cerr << "preferred-version: must include version "
-                << "0x" << config.version << std::endl;
+      std::cerr << "preferred-version: must include version " << std::hex
+                << "0x" << config.version << std::dec << std::endl;
       exit(EXIT_FAILURE);
     }
 
-    if (!config.other_versions.empty() &&
-        std::find(std::begin(config.other_versions),
-                  std::end(config.other_versions),
-                  config.version) == std::end(config.other_versions)) {
-      std::cerr << "other-versions: must include version "
-                << "0x" << config.version << std::endl;
+    if (!config.available_versions.empty() &&
+        std::find(std::begin(config.available_versions),
+                  std::end(config.available_versions),
+                  config.version) == std::end(config.available_versions)) {
+      std::cerr << "available-versions: must include version " << std::hex
+                << "0x" << config.version << std::dec << std::endl;
       exit(EXIT_FAILURE);
     }
   }
