@@ -273,7 +273,7 @@ void close_waitcb(struct ev_loop *loop, ev_timer *w, int revents) {
   auto s = h->server();
   auto conn = h->conn();
 
-  if (ngtcp2_conn_is_in_closing_period(conn)) {
+  if (ngtcp2_conn_in_closing_period(conn)) {
     if (!config.quiet) {
       std::cerr << "Closing Period is over" << std::endl;
     }
@@ -281,7 +281,7 @@ void close_waitcb(struct ev_loop *loop, ev_timer *w, int revents) {
     s->remove(h);
     return;
   }
-  if (ngtcp2_conn_is_in_draining_period(conn)) {
+  if (ngtcp2_conn_in_draining_period(conn)) {
     if (!config.quiet) {
       std::cerr << "Draining Period is over" << std::endl;
     }
@@ -433,14 +433,14 @@ int do_hp_mask(uint8_t *dest, const ngtcp2_crypto_cipher *hp,
 } // namespace
 
 namespace {
-int recv_crypto_data(ngtcp2_conn *conn, ngtcp2_crypto_level crypto_level,
-                     uint64_t offset, const uint8_t *data, size_t datalen,
-                     void *user_data) {
+int recv_crypto_data(ngtcp2_conn *conn,
+                     ngtcp2_encryption_level encryption_level, uint64_t offset,
+                     const uint8_t *data, size_t datalen, void *user_data) {
   if (!config.quiet && !config.no_quic_dump) {
-    debug::print_crypto_data(crypto_level, data, datalen);
+    debug::print_crypto_data(encryption_level, data, datalen);
   }
 
-  return ngtcp2_crypto_recv_crypto_data_cb(conn, crypto_level, offset, data,
+  return ngtcp2_crypto_recv_crypto_data_cb(conn, encryption_level, offset, data,
                                            datalen, user_data);
 }
 } // namespace
@@ -516,7 +516,7 @@ int stream_close(ngtcp2_conn *conn, uint32_t flags, int64_t stream_id,
 
 namespace {
 void rand(uint8_t *dest, size_t destlen, const ngtcp2_rand_ctx *rand_ctx) {
-  auto dis = std::uniform_int_distribution<uint8_t>(0, 255);
+  auto dis = std::uniform_int_distribution<uint8_t>();
   std::generate(dest, dest + destlen, [&dis]() { return dis(randgen); });
 }
 } // namespace
@@ -570,10 +570,43 @@ int update_key(ngtcp2_conn *conn, uint8_t *rx_secret, uint8_t *tx_secret,
 
 namespace {
 int path_validation(ngtcp2_conn *conn, uint32_t flags, const ngtcp2_path *path,
+                    const ngtcp2_path *old_path,
                     ngtcp2_path_validation_result res, void *user_data) {
   if (!config.quiet) {
     debug::path_validation(path, res);
   }
+
+  if (res != NGTCP2_PATH_VALIDATION_RESULT_SUCCESS ||
+      !(flags & NGTCP2_PATH_VALIDATION_FLAG_NEW_TOKEN)) {
+    return 0;
+  }
+
+  std::array<uint8_t, NGTCP2_CRYPTO_MAX_REGULAR_TOKENLEN> token;
+  auto t = std::chrono::duration_cast<std::chrono::nanoseconds>(
+               std::chrono::system_clock::now().time_since_epoch())
+               .count();
+
+  auto tokenlen = ngtcp2_crypto_generate_regular_token(
+      token.data(), config.static_secret.data(), config.static_secret.size(),
+      path->remote.addr, path->remote.addrlen, t);
+  if (tokenlen < 0) {
+    if (!config.quiet) {
+      std::cerr << "Unable to generate token" << std::endl;
+    }
+
+    return 0;
+  }
+
+  if (auto rv = ngtcp2_conn_submit_new_token(conn, token.data(), tokenlen);
+      rv != 0) {
+    if (!config.quiet) {
+      std::cerr << "ngtcp2_conn_submit_new_token: " << ngtcp2_strerror(rv)
+                << std::endl;
+    }
+
+    return NGTCP2_ERR_CALLBACK_FAILURE;
+  }
+
   return 0;
 }
 } // namespace
@@ -671,8 +704,9 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
   ngtcp2_settings settings;
   ngtcp2_settings_default(&settings);
   settings.log_printf = config.quiet ? nullptr : debug::log_printf;
-  settings.initial_ts = util::timestamp(loop_);
-  settings.token = ngtcp2_vec{const_cast<uint8_t *>(token), tokenlen};
+  settings.initial_ts = util::timestamp();
+  settings.token = token;
+  settings.tokenlen = tokenlen;
   settings.cc_algo = config.cc_algo;
   settings.initial_rtt = config.initial_rtt;
   settings.max_window = config.max_window;
@@ -695,16 +729,21 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
                 << strerror(errno) << std::endl;
       return -1;
     }
-    settings.qlog.write = ::write_qlog;
-    settings.qlog.odcid = *scid;
+    settings.qlog_write = ::write_qlog;
   }
   if (!config.preferred_versions.empty()) {
     settings.preferred_versions = config.preferred_versions.data();
     settings.preferred_versionslen = config.preferred_versions.size();
   }
-  if (!config.other_versions.empty()) {
-    settings.other_versions = config.other_versions.data();
-    settings.other_versionslen = config.other_versions.size();
+  if (!config.available_versions.empty()) {
+    settings.available_versions = config.available_versions.data();
+    settings.available_versionslen = config.available_versions.size();
+  }
+  if (config.initial_pkt_num == UINT32_MAX) {
+    auto dis = std::uniform_int_distribution<uint32_t>(0, INT32_MAX);
+    settings.initial_pkt_num = dis(randgen);
+  } else {
+    settings.initial_pkt_num = config.initial_pkt_num;
   }
 
   ngtcp2_transport_params params;
@@ -728,6 +767,8 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
     params.original_dcid = *scid;
   }
 
+  params.original_dcid_present = 1;
+
   if (util::generate_secure_random(params.stateless_reset_token,
                                    sizeof(params.stateless_reset_token)) != 0) {
     std::cerr << "Could not generate stateless reset token" << std::endl;
@@ -735,29 +776,28 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
   }
 
   if (config.preferred_ipv4_addr.len || config.preferred_ipv6_addr.len) {
-    params.preferred_address_present = 1;
+    params.preferred_addr_present = 1;
 
     if (config.preferred_ipv4_addr.len) {
-      params.preferred_address.ipv4 = config.preferred_ipv4_addr.su.in;
-      params.preferred_address.ipv4_present = 1;
+      params.preferred_addr.ipv4 = config.preferred_ipv4_addr.su.in;
+      params.preferred_addr.ipv4_present = 1;
     }
 
     if (config.preferred_ipv6_addr.len) {
-      params.preferred_address.ipv6 = config.preferred_ipv6_addr.su.in6;
-      params.preferred_address.ipv6_present = 1;
+      params.preferred_addr.ipv6 = config.preferred_ipv6_addr.su.in6;
+      params.preferred_addr.ipv6_present = 1;
     }
 
-    auto &token = params.preferred_address.stateless_reset_token;
+    auto &token = params.preferred_addr.stateless_reset_token;
     if (util::generate_secure_random(token, sizeof(token)) != 0) {
       std::cerr << "Could not generate preferred address stateless reset token"
                 << std::endl;
       return -1;
     }
 
-    params.preferred_address.cid.datalen = NGTCP2_SV_SCIDLEN;
-    if (util::generate_secure_random(params.preferred_address.cid.data,
-                                     params.preferred_address.cid.datalen) !=
-        0) {
+    params.preferred_addr.cid.datalen = NGTCP2_SV_SCIDLEN;
+    if (util::generate_secure_random(params.preferred_addr.cid.data,
+                                     params.preferred_addr.cid.datalen) != 0) {
       std::cerr << "Could not generate preferred address connection ID"
                 << std::endl;
       return -1;
@@ -798,7 +838,7 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
 
 int Handler::feed_data(const Endpoint &ep, const Address &local_addr,
                        const sockaddr *sa, socklen_t salen,
-                       const ngtcp2_pkt_info *pi, uint8_t *data,
+                       const ngtcp2_pkt_info *pi, const uint8_t *data,
                        size_t datalen) {
   auto path = ngtcp2_path{
       {
@@ -813,7 +853,7 @@ int Handler::feed_data(const Endpoint &ep, const Address &local_addr,
   };
 
   if (auto rv = ngtcp2_conn_read_pkt(conn_, &path, pi, data, datalen,
-                                     util::timestamp(loop_));
+                                     util::timestamp());
       rv != 0) {
     std::cerr << "ngtcp2_conn_read_pkt: " << ngtcp2_strerror(rv) << std::endl;
     switch (rv) {
@@ -826,14 +866,13 @@ int Handler::feed_data(const Endpoint &ep, const Address &local_addr,
       return NETWORK_ERR_DROP_CONN;
     case NGTCP2_ERR_CRYPTO:
       if (!last_error_.error_code) {
-        ngtcp2_connection_close_error_set_transport_error_tls_alert(
+        ngtcp2_ccerr_set_tls_alert(
             &last_error_, ngtcp2_conn_get_tls_alert(conn_), nullptr, 0);
       }
       break;
     default:
       if (!last_error_.error_code) {
-        ngtcp2_connection_close_error_set_transport_error_liberr(
-            &last_error_, rv, nullptr, 0);
+        ngtcp2_ccerr_set_liberr(&last_error_, rv, nullptr, 0);
       }
     }
     return handle_error();
@@ -844,7 +883,8 @@ int Handler::feed_data(const Endpoint &ep, const Address &local_addr,
 
 int Handler::on_read(const Endpoint &ep, const Address &local_addr,
                      const sockaddr *sa, socklen_t salen,
-                     const ngtcp2_pkt_info *pi, uint8_t *data, size_t datalen) {
+                     const ngtcp2_pkt_info *pi, const uint8_t *data,
+                     size_t datalen) {
   if (auto rv = feed_data(ep, local_addr, sa, salen, pi, data, datalen);
       rv != 0) {
     return rv;
@@ -856,12 +896,11 @@ int Handler::on_read(const Endpoint &ep, const Address &local_addr,
 }
 
 int Handler::handle_expiry() {
-  auto now = util::timestamp(loop_);
+  auto now = util::timestamp();
   if (auto rv = ngtcp2_conn_handle_expiry(conn_, now); rv != 0) {
     std::cerr << "ngtcp2_conn_handle_expiry: " << ngtcp2_strerror(rv)
               << std::endl;
-    ngtcp2_connection_close_error_set_transport_error_liberr(&last_error_, rv,
-                                                             nullptr, 0);
+    ngtcp2_ccerr_set_liberr(&last_error_, rv, nullptr, 0);
     return handle_error();
   }
 
@@ -869,8 +908,8 @@ int Handler::handle_expiry() {
 }
 
 int Handler::on_write() {
-  if (ngtcp2_conn_is_in_closing_period(conn_) ||
-      ngtcp2_conn_is_in_draining_period(conn_)) {
+  if (ngtcp2_conn_in_closing_period(conn_) ||
+      ngtcp2_conn_in_draining_period(conn_)) {
     return 0;
   }
 
@@ -884,6 +923,8 @@ int Handler::on_write() {
     }
   }
 
+  ev_io_stop(loop_, &wev_);
+
   if (auto rv = write_streams(); rv != 0) {
     return rv;
   }
@@ -905,7 +946,7 @@ int Handler::write_streams() {
   uint8_t *bufpos = tx_.data.get();
   ngtcp2_pkt_info pi;
   size_t gso_size = 0;
-  auto ts = util::timestamp(loop_);
+  auto ts = util::timestamp();
 
   ngtcp2_path_storage_zero(&ps);
   ngtcp2_path_storage_zero(&prev_ps);
@@ -953,8 +994,7 @@ int Handler::write_streams() {
 
       std::cerr << "ngtcp2_conn_writev_stream: " << ngtcp2_strerror(nwrite)
                 << std::endl;
-      ngtcp2_connection_close_error_set_transport_error_liberr(
-          &last_error_, nwrite, nullptr, 0);
+      ngtcp2_ccerr_set_liberr(&last_error_, nwrite, nullptr, 0);
       return handle_error();
     } else if (ndatalen >= 0) {
       stream->respbuf.pos += ndatalen;
@@ -979,13 +1019,9 @@ int Handler::write_streams() {
                           data + nsent, datalen - nsent, gso_size);
 
           start_wev_endpoint(ep);
-          ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-          return 0;
         }
       }
 
-      ev_io_stop(loop_, &wev_);
-
       // We are congestion limited.
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
       return 0;
@@ -1032,9 +1068,9 @@ int Handler::write_streams() {
 
           on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, data,
                           nwrite, 0);
-        }
 
-        start_wev_endpoint(ep);
+          start_wev_endpoint(ep);
+        }
       }
 
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
@@ -1054,9 +1090,10 @@ int Handler::write_streams() {
 
         on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, data + nsent,
                         datalen - nsent, gso_size);
+
+        start_wev_endpoint(ep);
       }
 
-      start_wev_endpoint(ep);
       ngtcp2_conn_update_pkt_tx_time(conn_, ts);
       return 0;
     }
@@ -1154,8 +1191,8 @@ void Handler::start_draining_period() {
 }
 
 int Handler::start_closing_period() {
-  if (!conn_ || ngtcp2_conn_is_in_closing_period(conn_) ||
-      ngtcp2_conn_is_in_draining_period(conn_)) {
+  if (!conn_ || ngtcp2_conn_in_closing_period(conn_) ||
+      ngtcp2_conn_in_draining_period(conn_)) {
     return 0;
   }
 
@@ -1180,7 +1217,7 @@ int Handler::start_closing_period() {
   ngtcp2_pkt_info pi;
   auto n = ngtcp2_conn_write_connection_close(
       conn_, &ps.path, &pi, conn_closebuf_->wpos(), conn_closebuf_->left(),
-      &last_error_, util::timestamp(loop_));
+      &last_error_, util::timestamp());
   if (n < 0) {
     std::cerr << "ngtcp2_conn_write_connection_close: " << ngtcp2_strerror(n)
               << std::endl;
@@ -1197,8 +1234,7 @@ int Handler::start_closing_period() {
 }
 
 int Handler::handle_error() {
-  if (last_error_.type ==
-      NGTCP2_CONNECTION_CLOSE_ERROR_CODE_TYPE_TRANSPORT_IDLE_CLOSE) {
+  if (last_error_.type == NGTCP2_CCERR_TYPE_IDLE_CLOSE) {
     return -1;
   }
 
@@ -1206,7 +1242,7 @@ int Handler::handle_error() {
     return -1;
   }
 
-  if (ngtcp2_conn_is_in_draining_period(conn_)) {
+  if (ngtcp2_conn_in_draining_period(conn_)) {
     return NETWORK_ERR_CLOSE_WAIT;
   }
 
@@ -1224,7 +1260,7 @@ int Handler::send_conn_close() {
 
   assert(conn_closebuf_ && conn_closebuf_->size());
   assert(conn_);
-  assert(!ngtcp2_conn_is_in_draining_period(conn_));
+  assert(!ngtcp2_conn_in_draining_period(conn_));
 
   auto path = ngtcp2_conn_get_path(conn_);
 
@@ -1235,12 +1271,13 @@ int Handler::send_conn_close() {
 
 void Handler::update_timer() {
   auto expiry = ngtcp2_conn_get_expiry(conn_);
-  auto now = util::timestamp(loop_);
+  auto now = util::timestamp();
 
   if (expiry <= now) {
     if (!config.quiet) {
       auto t = static_cast<ev_tstamp>(now - expiry) / NGTCP2_SECONDS;
-      std::cerr << "Timer has already expired: " << t << "s" << std::endl;
+      std::cerr << "Timer has already expired: " << std::fixed << t << "s"
+                << std::defaultfloat << std::endl;
     }
 
     ev_feed_event(loop_, &timer_, EV_TIMER);
@@ -1314,13 +1351,12 @@ int Handler::recv_stream_data(uint32_t flags, int64_t stream_id,
         http_parser_execute(&stream->htp, &htp_settings,
                             reinterpret_cast<const char *>(data), datalen);
     if (nread != datalen) {
-      if (auto rv = ngtcp2_conn_shutdown_stream(conn_, stream_id,
+      if (auto rv = ngtcp2_conn_shutdown_stream(conn_, 0, stream_id,
                                                 /* app error code */ 1);
           rv != 0) {
         std::cerr << "ngtcp2_conn_shutdown_stream: " << ngtcp2_strerror(rv)
                   << std::endl;
-        ngtcp2_connection_close_error_set_transport_error_liberr(
-            &last_error_, NGTCP2_ERR_INTERNAL, nullptr, 0);
+        ngtcp2_ccerr_set_liberr(&last_error_, NGTCP2_ERR_INTERNAL, nullptr, 0);
         return -1;
       }
     }
@@ -1394,7 +1430,7 @@ int Handler::on_stream_close(int64_t stream_id, uint64_t app_error_code) {
 }
 
 void Handler::shutdown_read(int64_t stream_id, int app_error_code) {
-  ngtcp2_conn_shutdown_stream_read(conn_, stream_id, app_error_code);
+  ngtcp2_conn_shutdown_stream_read(conn_, 0, stream_id, app_error_code);
 }
 
 void Handler::add_sendq(Stream *stream) { sendq_.emplace(stream); }
@@ -1651,8 +1687,7 @@ int Server::init(const char *addr, const char *port) {
 
 int Server::on_read(Endpoint &ep) {
   sockaddr_union su;
-  std::array<uint8_t, 64_k> buf;
-  ngtcp2_pkt_hd hd;
+  std::array<uint8_t, 64_k - 1> buf;
   size_t pktcnt = 0;
   ngtcp2_pkt_info pi;
 
@@ -1665,8 +1700,9 @@ int Server::on_read(Endpoint &ep) {
   msg.msg_iov = &msg_iov;
   msg.msg_iovlen = 1;
 
-  uint8_t
-      msg_ctrl[CMSG_SPACE(sizeof(uint8_t)) + CMSG_SPACE(sizeof(in6_pktinfo))];
+  uint8_t msg_ctrl[CMSG_SPACE(sizeof(uint8_t)) +
+                   CMSG_SPACE(sizeof(in6_pktinfo)) +
+                   CMSG_SPACE(sizeof(uint16_t))];
   msg.msg_control = msg_ctrl;
 
   for (; pktcnt < 10;) {
@@ -1681,206 +1717,214 @@ int Server::on_read(Endpoint &ep) {
       return 0;
     }
 
-    ++pktcnt;
-
     pi.ecn = msghdr_get_ecn(&msg, su.storage.ss_family);
     auto local_addr = msghdr_get_local_addr(&msg, su.storage.ss_family);
     if (!local_addr) {
+      ++pktcnt;
       std::cerr << "Unable to obtain local address" << std::endl;
       continue;
     }
 
+    auto gso_size = msghdr_get_udp_gro(&msg);
+    if (gso_size == 0) {
+      gso_size = static_cast<size_t>(nread);
+    }
+
     set_port(*local_addr, ep.addr);
 
-    if (!config.quiet) {
-      std::array<char, IF_NAMESIZE> ifname;
-      std::cerr << "Received packet: local="
-                << util::straddr(&local_addr->su.sa, local_addr->len)
-                << " remote=" << util::straddr(&su.sa, msg.msg_namelen)
-                << " if=" << if_indextoname(local_addr->ifindex, ifname.data())
-                << " ecn=0x" << std::hex << pi.ecn << std::dec << " " << nread
-                << " bytes" << std::endl;
-    }
+    auto data = buf.data();
+
+    for (;;) {
+      auto datalen = std::min(static_cast<size_t>(nread), gso_size);
+
+      ++pktcnt;
 
-    if (debug::packet_lost(config.rx_loss_prob)) {
       if (!config.quiet) {
-        std::cerr << "** Simulated incoming packet loss **" << std::endl;
+        std::array<char, IF_NAMESIZE> ifname;
+        std::cerr << "Received packet: local="
+                  << util::straddr(&local_addr->su.sa, local_addr->len)
+                  << " remote=" << util::straddr(&su.sa, msg.msg_namelen)
+                  << " if="
+                  << if_indextoname(local_addr->ifindex, ifname.data())
+                  << " ecn=0x" << std::hex << pi.ecn << std::dec << " "
+                  << datalen << " bytes" << std::endl;
       }
-      continue;
-    }
 
-    if (nread == 0) {
-      continue;
-    }
-
-    ngtcp2_version_cid vc;
-
-    switch (auto rv = ngtcp2_pkt_decode_version_cid(&vc, buf.data(), nread,
-                                                    NGTCP2_SV_SCIDLEN);
-            rv) {
-    case 0:
-      break;
-    case NGTCP2_ERR_VERSION_NEGOTIATION:
-      send_version_negotiation(vc.version, vc.scid, vc.scidlen, vc.dcid,
-                               vc.dcidlen, ep, *local_addr, &su.sa,
-                               msg.msg_namelen);
-      continue;
-    default:
-      std::cerr << "Could not decode version and CID from QUIC packet header: "
-                << ngtcp2_strerror(rv) << std::endl;
-      continue;
-    }
-
-    auto dcid_key = util::make_cid_key(vc.dcid, vc.dcidlen);
-
-    auto handler_it = handlers_.find(dcid_key);
-    if (handler_it == std::end(handlers_)) {
-      switch (auto rv = ngtcp2_accept(&hd, buf.data(), nread); rv) {
-      case 0:
-        break;
-      case NGTCP2_ERR_RETRY:
-        send_retry(&hd, ep, *local_addr, &su.sa, msg.msg_namelen, nread * 3);
-        continue;
-      default:
+      if (debug::packet_lost(config.rx_loss_prob)) {
         if (!config.quiet) {
-          std::cerr << "Unexpected packet received: length=" << nread
-                    << std::endl;
+          std::cerr << "** Simulated incoming packet loss **" << std::endl;
         }
-        continue;
+      } else {
+        read_pkt(ep, *local_addr, &su.sa, msg.msg_namelen, &pi, data, datalen);
       }
 
-      ngtcp2_cid ocid;
-      ngtcp2_cid *pocid = nullptr;
-
-      assert(hd.type == NGTCP2_PKT_INITIAL);
-
-      if (config.validate_addr || hd.token.len) {
-        std::cerr << "Perform stateless address validation" << std::endl;
-        if (hd.token.len == 0) {
-          send_retry(&hd, ep, *local_addr, &su.sa, msg.msg_namelen, nread * 3);
-          continue;
-        }
-
-        if (hd.token.base[0] != NGTCP2_CRYPTO_TOKEN_MAGIC_RETRY &&
-            hd.dcid.datalen < NGTCP2_MIN_INITIAL_DCIDLEN) {
-          send_stateless_connection_close(&hd, ep, *local_addr, &su.sa,
-                                          msg.msg_namelen);
-          continue;
-        }
-
-        switch (hd.token.base[0]) {
-        case NGTCP2_CRYPTO_TOKEN_MAGIC_RETRY:
-          if (verify_retry_token(&ocid, &hd, &su.sa, msg.msg_namelen) != 0) {
-            send_stateless_connection_close(&hd, ep, *local_addr, &su.sa,
-                                            msg.msg_namelen);
-            continue;
-          }
-          pocid = &ocid;
-          break;
-        case NGTCP2_CRYPTO_TOKEN_MAGIC_REGULAR:
-          if (verify_token(&hd, &su.sa, msg.msg_namelen) != 0) {
-            if (config.validate_addr) {
-              send_retry(&hd, ep, *local_addr, &su.sa, msg.msg_namelen,
-                         nread * 3);
-              continue;
-            }
-
-            hd.token.base = nullptr;
-            hd.token.len = 0;
-          }
-          break;
-        default:
-          if (!config.quiet) {
-            std::cerr << "Ignore unrecognized token" << std::endl;
-          }
-          if (config.validate_addr) {
-            send_retry(&hd, ep, *local_addr, &su.sa, msg.msg_namelen,
-                       nread * 3);
-            continue;
-          }
-
-          hd.token.base = nullptr;
-          hd.token.len = 0;
-          break;
-        }
-      }
-
-      auto h = std::make_unique<Handler>(loop_, this);
-      if (h->init(ep, *local_addr, &su.sa, msg.msg_namelen, &hd.scid, &hd.dcid,
-                  pocid, hd.token.base, hd.token.len, hd.version,
-                  tls_ctx_) != 0) {
-        continue;
-      }
-
-      switch (h->on_read(ep, *local_addr, &su.sa, msg.msg_namelen, &pi,
-                         buf.data(), nread)) {
-      case 0:
+      nread -= datalen;
+      if (nread == 0) {
         break;
-      case NETWORK_ERR_RETRY:
-        send_retry(&hd, ep, *local_addr, &su.sa, msg.msg_namelen, nread * 3);
-        continue;
-      default:
-        continue;
       }
 
-      switch (h->on_write()) {
-      case 0:
-        break;
-      default:
-        continue;
-      }
-
-      std::array<ngtcp2_cid, 2> scids;
-      auto conn = h->conn();
-
-      auto num_scid = ngtcp2_conn_get_num_scid(conn);
-
-      assert(num_scid <= scids.size());
-
-      ngtcp2_conn_get_scid(conn, scids.data());
-
-      for (size_t i = 0; i < num_scid; ++i) {
-        handlers_.emplace(util::make_cid_key(&scids[i]), h.get());
-      }
-
-      handlers_.emplace(dcid_key, h.get());
-
-      h.release();
-
-      continue;
+      data += datalen;
     }
-
-    auto h = (*handler_it).second;
-    auto conn = h->conn();
-    if (ngtcp2_conn_is_in_closing_period(conn)) {
-      // TODO do exponential backoff.
-      switch (h->send_conn_close()) {
-      case 0:
-        break;
-      default:
-        remove(h);
-      }
-      continue;
-    }
-    if (ngtcp2_conn_is_in_draining_period(conn)) {
-      continue;
-    }
-
-    if (auto rv = h->on_read(ep, *local_addr, &su.sa, msg.msg_namelen, &pi,
-                             buf.data(), nread);
-        rv != 0) {
-      if (rv != NETWORK_ERR_CLOSE_WAIT) {
-        remove(h);
-      }
-      continue;
-    }
-
-    h->signal_write();
   }
 
   return 0;
 }
 
+void Server::read_pkt(Endpoint &ep, const Address &local_addr,
+                      const sockaddr *sa, socklen_t salen,
+                      const ngtcp2_pkt_info *pi, const uint8_t *data,
+                      size_t datalen) {
+  if (datalen == 0) {
+    return;
+  }
+
+  ngtcp2_version_cid vc;
+
+  switch (auto rv = ngtcp2_pkt_decode_version_cid(&vc, data, datalen,
+                                                  NGTCP2_SV_SCIDLEN);
+          rv) {
+  case 0:
+    break;
+  case NGTCP2_ERR_VERSION_NEGOTIATION:
+    send_version_negotiation(vc.version, vc.scid, vc.scidlen, vc.dcid,
+                             vc.dcidlen, ep, local_addr, sa, salen);
+    return;
+  default:
+    std::cerr << "Could not decode version and CID from QUIC packet header: "
+              << ngtcp2_strerror(rv) << std::endl;
+    return;
+  }
+
+  auto dcid_key = util::make_cid_key(vc.dcid, vc.dcidlen);
+
+  auto handler_it = handlers_.find(dcid_key);
+  if (handler_it == std::end(handlers_)) {
+    ngtcp2_pkt_hd hd;
+
+    if (auto rv = ngtcp2_accept(&hd, data, datalen); rv != 0) {
+      if (!config.quiet) {
+        std::cerr << "Unexpected packet received: length=" << datalen
+                  << std::endl;
+      }
+      return;
+    }
+
+    ngtcp2_cid ocid;
+    ngtcp2_cid *pocid = nullptr;
+
+    assert(hd.type == NGTCP2_PKT_INITIAL);
+
+    if (config.validate_addr || hd.tokenlen) {
+      std::cerr << "Perform stateless address validation" << std::endl;
+      if (hd.tokenlen == 0) {
+        send_retry(&hd, ep, local_addr, sa, salen, datalen * 3);
+        return;
+      }
+
+      if (hd.token[0] != NGTCP2_CRYPTO_TOKEN_MAGIC_RETRY &&
+          hd.dcid.datalen < NGTCP2_MIN_INITIAL_DCIDLEN) {
+        send_stateless_connection_close(&hd, ep, local_addr, sa, salen);
+        return;
+      }
+
+      switch (hd.token[0]) {
+      case NGTCP2_CRYPTO_TOKEN_MAGIC_RETRY:
+        if (verify_retry_token(&ocid, &hd, sa, salen) != 0) {
+          send_stateless_connection_close(&hd, ep, local_addr, sa, salen);
+          return;
+        }
+        pocid = &ocid;
+        break;
+      case NGTCP2_CRYPTO_TOKEN_MAGIC_REGULAR:
+        if (verify_token(&hd, sa, salen) != 0) {
+          if (config.validate_addr) {
+            send_retry(&hd, ep, local_addr, sa, salen, datalen * 3);
+            return;
+          }
+
+          hd.token = nullptr;
+          hd.tokenlen = 0;
+        }
+        break;
+      default:
+        if (!config.quiet) {
+          std::cerr << "Ignore unrecognized token" << std::endl;
+        }
+        if (config.validate_addr) {
+          send_retry(&hd, ep, local_addr, sa, salen, datalen * 3);
+          return;
+        }
+
+        hd.token = nullptr;
+        hd.tokenlen = 0;
+        break;
+      }
+    }
+
+    auto h = std::make_unique<Handler>(loop_, this);
+    if (h->init(ep, local_addr, sa, salen, &hd.scid, &hd.dcid, pocid, hd.token,
+                hd.tokenlen, hd.version, tls_ctx_) != 0) {
+      return;
+    }
+
+    switch (h->on_read(ep, local_addr, sa, salen, pi, data, datalen)) {
+    case 0:
+      break;
+    case NETWORK_ERR_RETRY:
+      send_retry(&hd, ep, local_addr, sa, salen, datalen * 3);
+      return;
+    default:
+      return;
+    }
+
+    if (h->on_write() != 0) {
+      return;
+    }
+
+    std::array<ngtcp2_cid, 2> scids;
+    auto conn = h->conn();
+
+    auto num_scid = ngtcp2_conn_get_scid(conn, nullptr);
+
+    assert(num_scid <= scids.size());
+
+    ngtcp2_conn_get_scid(conn, scids.data());
+
+    for (size_t i = 0; i < num_scid; ++i) {
+      associate_cid(&scids[i], h.get());
+    }
+
+    handlers_.emplace(dcid_key, h.get());
+
+    h.release();
+
+    return;
+  }
+
+  auto h = (*handler_it).second;
+  auto conn = h->conn();
+  if (ngtcp2_conn_in_closing_period(conn)) {
+    // TODO do exponential backoff.
+    if (h->send_conn_close() != 0) {
+      remove(h);
+    }
+    return;
+  }
+  if (ngtcp2_conn_in_draining_period(conn)) {
+    return;
+  }
+
+  if (auto rv = h->on_read(ep, local_addr, sa, salen, pi, data, datalen);
+      rv != 0) {
+    if (rv != NETWORK_ERR_CLOSE_WAIT) {
+      remove(h);
+    }
+    return;
+  }
+
+  h->signal_write();
+}
+
 namespace {
 uint32_t generate_reserved_version(const sockaddr *sa, socklen_t salen,
                                    uint32_t version) {
@@ -1925,9 +1969,7 @@ int Server::send_version_negotiation(uint32_t version, const uint8_t *dcid,
   }
 
   auto nwrite = ngtcp2_pkt_write_version_negotiation(
-      buf.wpos(), buf.left(),
-      std::uniform_int_distribution<uint8_t>(
-          0, std::numeric_limits<uint8_t>::max())(randgen),
+      buf.wpos(), buf.left(), std::uniform_int_distribution<uint8_t>()(randgen),
       dcid, dcidlen, scid, scidlen, sv.data(), p - std::begin(sv));
   if (nwrite < 0) {
     std::cerr << "ngtcp2_pkt_write_version_negotiation: "
@@ -2076,7 +2118,7 @@ int Server::verify_retry_token(ngtcp2_cid *ocid, const ngtcp2_pkt_hd *hd,
   if (!config.quiet) {
     std::cerr << "Verifying Retry token from [" << host.data()
               << "]:" << port.data() << std::endl;
-    util::hexdump(stderr, hd->token.base, hd->token.len);
+    util::hexdump(stderr, hd->token, hd->tokenlen);
   }
 
   auto t = std::chrono::duration_cast<std::chrono::nanoseconds>(
@@ -2084,7 +2126,7 @@ int Server::verify_retry_token(ngtcp2_cid *ocid, const ngtcp2_pkt_hd *hd,
                .count();
 
   if (ngtcp2_crypto_verify_retry_token(
-          ocid, hd->token.base, hd->token.len, config.static_secret.data(),
+          ocid, hd->token, hd->tokenlen, config.static_secret.data(),
           config.static_secret.size(), hd->version, sa, salen, &hd->dcid,
           10 * NGTCP2_SECONDS, t) != 0) {
     std::cerr << "Could not verify Retry token" << std::endl;
@@ -2114,14 +2156,14 @@ int Server::verify_token(const ngtcp2_pkt_hd *hd, const sockaddr *sa,
   if (!config.quiet) {
     std::cerr << "Verifying token from [" << host.data() << "]:" << port.data()
               << std::endl;
-    util::hexdump(stderr, hd->token.base, hd->token.len);
+    util::hexdump(stderr, hd->token, hd->tokenlen);
   }
 
   auto t = std::chrono::duration_cast<std::chrono::nanoseconds>(
                std::chrono::system_clock::now().time_since_epoch())
                .count();
 
-  if (ngtcp2_crypto_verify_regular_token(hd->token.base, hd->token.len,
+  if (ngtcp2_crypto_verify_regular_token(hd->token, hd->tokenlen,
                                          config.static_secret.data(),
                                          config.static_secret.size(), sa, salen,
                                          3600 * NGTCP2_SECONDS, t) != 0) {
@@ -2303,14 +2345,13 @@ void Server::dissociate_cid(const ngtcp2_cid *cid) {
 void Server::remove(const Handler *h) {
   auto conn = h->conn();
 
-  handlers_.erase(
-      util::make_cid_key(ngtcp2_conn_get_client_initial_dcid(conn)));
+  dissociate_cid(ngtcp2_conn_get_client_initial_dcid(conn));
 
-  std::vector<ngtcp2_cid> cids(ngtcp2_conn_get_num_scid(conn));
+  std::vector<ngtcp2_cid> cids(ngtcp2_conn_get_scid(conn, nullptr));
   ngtcp2_conn_get_scid(conn, cids.data());
 
   for (auto &cid : cids) {
-    handlers_.erase(util::make_cid_key(&cid));
+    dissociate_cid(&cid);
   }
 
   delete h;
@@ -2406,8 +2447,9 @@ void config_set_default(Config &config) {
   config.cc_algo = NGTCP2_CC_ALGO_CUBIC;
   config.initial_rtt = NGTCP2_DEFAULT_INITIAL_RTT;
   config.max_gso_dgrams = 64;
-  config.handshake_timeout = NGTCP2_DEFAULT_HANDSHAKE_TIMEOUT;
+  config.handshake_timeout = UINT64_MAX;
   config.ack_thresh = 2;
+  config.initial_pkt_num = UINT32_MAX;
 }
 } // namespace
 
@@ -2541,31 +2583,35 @@ Options:
               Default: )"
             << config.max_gso_dgrams << R"(
   --handshake-timeout=<DURATION>
-              Set the QUIC handshake timeout.
-              Default: )"
-            << util::format_duration(config.handshake_timeout) << R"(
+              Set  the  QUIC handshake  timeout.   It  defaults to  no
+              timeout.
   --preferred-versions=<HEX>[[,<HEX>]...]
               Specify  QUIC versions  in hex  string in  the order  of
               preference.  Server negotiates one  of those versions if
               client  initially  selects  a  less  preferred  version.
               These versions must be  supported by libngtcp2.  Instead
               of  specifying hex  string,  there  are special  aliases
-              available:  "v1"   indicates  QUIC  v1,   and  "v2draft"
-              indicates QUIC v2 draft.
-  --other-versions=<HEX>[[,<HEX>]...]
+              available: "v1"  indicates QUIC  v1, and  "v2" indicates
+              QUIC v2.
+  --available-versions=<HEX>[[,<HEX>]...]
               Specify QUIC  versions in  hex string  that are  sent in
-              other_versions  field  of version_information  transport
-              parameter.  This list can include a version which is not
-              supported  by  libngtcp2.   Instead  of  specifying  hex
-              string,  there  are   special  aliases  available:  "v1"
-              indicates  QUIC  v1,  and "v2draft"  indicates  QUIC  v2
-              draft.
+              available_versions    field    of    version_information
+              transport parameter.   This list  can include  a version
+              which  is  not  supported   by  libngtcp2.   Instead  of
+              specifying  hex   string,  there  are   special  aliases
+              available: "v1"  indicates QUIC  v1, and  "v2" indicates
+              QUIC v2.
   --no-pmtud  Disables Path MTU Discovery.
   --ack-thresh=<N>
               The minimum number of the received ACK eliciting packets
               that triggers immediate acknowledgement.
               Default: )"
             << config.ack_thresh << R"(
+  --initial-pkt-num=<N>
+              The initial packet  number that is used  for each packet
+              number space.  It  must be in range [0, (1  << 31) - 1],
+              inclusive.   By default,  the initial  packet number  is
+              chosen randomly.
   -h, --help  Display this help and exit.
 
 ---
@@ -2627,9 +2673,10 @@ int main(int argc, char **argv) {
         {"max-gso-dgrams", required_argument, &flag, 25},
         {"handshake-timeout", required_argument, &flag, 26},
         {"preferred-versions", required_argument, &flag, 27},
-        {"other-versions", required_argument, &flag, 28},
+        {"available-versions", required_argument, &flag, 28},
         {"no-pmtud", no_argument, &flag, 29},
         {"ack-thresh", required_argument, &flag, 30},
+        {"initial-pkt-num", required_argument, &flag, 31},
         {nullptr, 0, nullptr, 0}};
 
     auto optidx = 0;
@@ -2898,8 +2945,8 @@ int main(int argc, char **argv) {
             *it++ = NGTCP2_PROTO_VER_V1;
             continue;
           }
-          if (k == "v2draft"sv) {
-            *it++ = NGTCP2_PROTO_VER_V2_DRAFT;
+          if (k == "v2"sv) {
+            *it++ = NGTCP2_PROTO_VER_V2;
             continue;
           }
           auto rv = util::parse_version(k);
@@ -2918,23 +2965,23 @@ int main(int argc, char **argv) {
         break;
       }
       case 28: {
-        // --other-versions
+        // --available-versions
         auto l = util::split_str(optarg);
-        config.other_versions.resize(l.size());
-        auto it = std::begin(config.other_versions);
+        config.available_versions.resize(l.size());
+        auto it = std::begin(config.available_versions);
         for (const auto &k : l) {
           if (k == "v1"sv) {
             *it++ = NGTCP2_PROTO_VER_V1;
             continue;
           }
-          if (k == "v2draft"sv) {
-            *it++ = NGTCP2_PROTO_VER_V2_DRAFT;
+          if (k == "v2"sv) {
+            *it++ = NGTCP2_PROTO_VER_V2;
             continue;
           }
           auto rv = util::parse_version(k);
           if (!rv) {
-            std::cerr << "other-versions: invalid version " << std::quoted(k)
-                      << std::endl;
+            std::cerr << "available-versions: invalid version "
+                      << std::quoted(k) << std::endl;
             exit(EXIT_FAILURE);
           }
           *it++ = *rv;
@@ -2957,6 +3004,19 @@ int main(int argc, char **argv) {
           config.ack_thresh = *n;
         }
         break;
+      case 31:
+        // --initial-pkt-num
+        if (auto n = util::parse_uint(optarg); !n) {
+          std::cerr << "initial-pkt-num: invalid argument" << std::endl;
+          exit(EXIT_FAILURE);
+        } else if (*n > INT32_MAX) {
+          std::cerr << "initial-pkt-num: must not exceed (1 << 31) - 1"
+                    << std::endl;
+          exit(EXIT_FAILURE);
+        } else {
+          config.initial_pkt_num = static_cast<uint32_t>(*n);
+        }
+        break;
       }
       break;
     default:
